from torch import nn
from transformers import DistilBertModel


class TextEncoder(nn.Module):

    def __init__(self, model_name):
        super(TextEncoder, self).__init__()
        self.model = DistilBertModel.from_pretrained(model_name)

    def forward(self, input_ids, attention_mask):
        output = self.model(input_ids=input_ids, attention_mask=attention_mask)
        last_hidden_state = output.last_hidden_state
        return last_hidden_state.mean(dim=1)



import torch
import torchvision.transforms as transforms
import numpy as np
from collections import defaultdict

from tqdm.auto import tqdm
from io import BytesIO
from base64 import b64decode

np.random.seed(42)

from PIL import Image

def get_image_tranforms(image_size=(224, 224)):
    return transforms.Compose([
        transforms.Resize(image_size),
        transforms.Lambda(_grayscale_to_rgb),
        transforms.ToTensor()
    ])


def _grayscale_to_rgb(img):
    if img.mode != "RGB":
        return img.convert("RGB")
    return img


class COCODataset(torch.utils.data.Dataset):

    def __init__(self,
                 data,
                 tokenizer,
                 transforms,
                 image_key,
                 text_key,
                 shuffle_captions=True):

        captions = []
        for row in tqdm(data):
            captions.append(row[text_key])

        print("captions", len(captions), captions[0])
        encoded_captions = tokenizer(captions,
                                     padding=True,
                                     truncation=True,
                                     max_length=100)

        # group images, since we can have multiple captions for an image
        # we avoid having duplicate images in a single batch
        grouped_data = defaultdict(list)
        for idx, row in enumerate(tqdm(data)):
            grouped_data[row[image_key][0]].append({
                "input_ids":
                encoded_captions["input_ids"][idx],
                "attention_mask":
                encoded_captions["attention_mask"][idx]
            })

        self.data = list(grouped_data.items())
        print("data", len(self.data))
        self.transforms = transforms
        self.shuffle_captions = shuffle_captions

    def __getitem__(self, idx):
        image_str, encoded_texts = self.data[idx]
        image = Image.open(BytesIO(b64decode(image_str)))
        image = self.transforms(image)

        if self.shuffle_captions:
            encoded_text = np.random.choice(encoded_texts)
        else:
            encoded_text = encoded_texts[0]
        instance = {
            key: torch.tensor(value)
            for key, value in encoded_text.items()
        }
        instance["image"] = image

        return instance

    def __len__(self):
        return len(self.data)


class SBUDataset(torch.utils.data.Dataset):

    def __init__(self,
                 data,
                 tokenizer,
                 transforms,
                 image_key="image_path",
                 text_key="captions",
                 shuffle_captions=True):

        valid_indices = []
        captions = []
        for idx, row in enumerate(tqdm(data)):
            if row[image_key]:
                captions.append(row[text_key])
                valid_indices.append(idx)
        print("captions", len(captions), captions[0])
        self.encoded_captions = tokenizer(captions,
                                          padding=True,
                                          truncation=True,
                                          max_length=100)

        self.data = data.select(valid_indices)
        print("data", len(self.data),
              len(self.encoded_captions["input_ids"][0]))
        self.transforms = transforms
        self.shuffle_captions = shuffle_captions

    def __getitem__(self, idx):
        image = self.data[idx]["image"]
        image = self.transforms(image)

        instance = {
            key: torch.tensor(value[idx])
            for key, value in self.encoded_captions.items()
        }
        instance["image"] = image

        return instance

    def __len__(self):
        return len(self.data)
    

# coco + textcap + sbucaptions
class CombinedDataset(torch.utils.data.Dataset):

    def __init__(self,
                 data,
                 tokenizer,
                 transforms,
                 shuffle_captions=True):

        self.tokenizer = tokenizer

        self.data = data
        self.transforms = transforms
        self.shuffle_captions = shuffle_captions

    def __getitem__(self, idx):
        row = self.data[idx]
        image, captions = row["image"], row["caption"]

        if self.shuffle_captions:
            caption = np.random.choice(captions)
        else:
            caption = captions[0]
        encoded_caption = self.tokenizer(caption,
                                    padding="max_length",
                                    truncation=True,
                                    max_length=100)
        instance = {
            key: torch.tensor(value)
            for key, value in encoded_caption.items()
        }
        instance["image"] = self.transforms(image)

        return instance

    def __len__(self):
        return len(self.data)



import torch
from torch.utils.data import DataLoader
from torch.cuda.amp import GradScaler, autocast
import random
import numpy as np
from tqdm.auto import tqdm
from torchinfo import summary
import webdataset as wds

from simple_clip import CLIP, clip_loss

from simple_clip.utils import accuracy, get_dataset, get_image_encoder, get_text_encoder
from simple_clip.custom_datasets.clip_datasets import get_image_tranforms
from simple_clip.imagenet_eval import ImageNetValidation

SEED = 42

random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed(SEED)


def train_clip(args):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    image_encoder = get_image_encoder(args.image_encoder_name)
    text_encoder = get_text_encoder(args.text_encoder_name)
    model = CLIP(image_encoder, text_encoder)

    img = torch.rand(1, 3, args.image_size, args.image_size)
    txt = torch.randint(high=20000, size=(1, 100))
    att_mask = torch.randint(high=1, size=(1, 100))
    summary(model, input_data=[img, txt, att_mask])

    if args.ckpt_path:
        model_state = torch.load(args.ckpt_path)
        model.load_state_dict(model_state)
    model = model.to(device)

    optimizer = torch.optim.Adam(list(model.parameters()),
                                 lr=args.learning_rate,
                                 weight_decay=args.weight_decay)

    transforms_inference = get_image_tranforms(
        (args.image_size, args.image_size))
    ds = get_dataset(args.dataset_name,
                     args.dataset_path,
                     transforms=transforms_inference)
    if args.dataset_name == "yfcc7m":
        train_loader = wds.WebLoader(ds, num_workers=2, batch_size=args.batch_size)
        # hardcoded for num images = 7329280 :/
        steps_per_epcoch = 7329280 // args.batch_size
    else:
        train_loader = DataLoader(ds,
                              batch_size=args.batch_size,
                              num_workers=12,
                              drop_last=True,
                              shuffle=True)
        steps_per_epcoch = len(train_loader)

    tmax = args.num_epochs * steps_per_epcoch + steps_per_epcoch // 4
    lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=(tmax), eta_min=1e-8)

    # Always use coco val dataset for standardization purposes
    ds_val = get_dataset("coco",
                         args.dataset_path,
                         transforms=transforms_inference,
                         split="validation",
                         shuffle_captions=False)
    val_loader = DataLoader(ds_val,
                            batch_size=min(args.batch_size, 256),
                            num_workers=4)
    
    if args.imagenet_eval:
        imgnet_val = ImageNetValidation(transforms_inference)

    scaler = GradScaler(enabled=args.fp16_precision)

    global_step = 0
    total_loss = 0.0
    for epoch in range(args.num_epochs):
        epoch_loss = 0.0
        epoch_images_loss = 0.0
        epoch_texts_loss = 0.0
        progress_bar = tqdm(train_loader, desc=f"Epoch {epoch+1}/{args.num_epochs}", total=steps_per_epcoch)
        for step, (batch) in enumerate(progress_bar):
            batch = {k: v.to(device) for k, v in batch.items() if k in ['input_ids', 'attention_mask', 'image']}

            with autocast(enabled=args.fp16_precision):
                logits = model(**batch)
                loss_images, loss_texts = clip_loss(logits)
                loss = (loss_images + loss_texts) / 2

            optimizer.zero_grad()
            scaler.scale(loss).backward()
            scaler.step(optimizer)
            scaler.update()

            total_loss += loss.item()
            epoch_loss += loss.item()
            avg_loss = total_loss / (global_step + 1)
            ep_loss = epoch_loss / (step + 1)

            epoch_images_loss += loss_images.item()
            ep_images_loss = epoch_images_loss / (step + 1)
            epoch_texts_loss += loss_texts.item()
            ep_texts_loss = epoch_texts_loss / (step + 1)

            current_lr = optimizer.param_groups[0]['lr']
            progress_bar.set_description(
                f"Epoch {epoch+1}/{args.num_epochs} | "
                f"Step {global_step+1} | "
                f"Epoch Loss: {ep_loss:.4f} |"
                f"Total Loss: {avg_loss:.4f} |"
                f"Images Loss: {ep_images_loss:.6f} |"
                f"Texts Loss: {ep_texts_loss:.6f} |"
                f"Lr: {current_lr:.8f}")

            global_step += 1
            if global_step % args.log_every_n_steps == 0:
                batch_size = logits.shape[0]
                labels = torch.arange(batch_size).to(logits.device)
                top1, top5 = accuracy(logits, labels, topk=(1, 5))
                print('acc/top1 images', top1[0].item())
                print('acc/top5 images', top5[0].item())
                top1, top5 = accuracy(logits.t(), labels, topk=(1, 5))
                print('acc/top1 texts', top1[0].item())
                print('acc/top5 texts', top5[0].item())

                torch.save(model.state_dict(),
                           f"{args.save_model_dir}/clip_model.pth")

            if global_step % (args.log_every_n_steps * 10) == 0:
                validate(model, val_loader, device)
            
            if args.imagenet_eval and global_step % args.imagenet_eval_steps == 0:
                imgnet_val.evaluate(model)

            lr_scheduler.step()


def validate(model, val_loader, device):
    top1_acc_images = []
    top5_acc_images = []
    top1_acc_texts = []
    top5_acc_texts = []
    model.eval()
    with torch.no_grad():
        for batch in tqdm(val_loader):
            batch = {k: v.to(device) for k, v in batch.items()}

            logits = model(**batch)

            labels = torch.arange(logits.size(0)).to(device)
            top1, top5 = accuracy(logits, labels, topk=(1, 5))
            top1_acc_images.append(top1[0].item())
            top5_acc_images.append(top5[0].item())
            top1, top5 = accuracy(logits.t(), labels, topk=(1, 5))
            top1_acc_texts.append(top1[0].item())
            top5_acc_texts.append(top5[0].item())
    print("#" * 100)
    print('eval acc/top1 images', np.mean(top1_acc_images))
    print('eval acc/top5 images', np.mean(top5_acc_images))
    print('eval acc/top1 texts', np.mean(top1_acc_texts))
    print('eval acc/top5 texts', np.mean(top5_acc_texts))



import argparse

from train import train_clip

parser = argparse.ArgumentParser(description='CLIP')
parser.add_argument('--dataset_path',
                    default='./data',
                    help='Path where datasets will be saved')
parser.add_argument('--dataset_name',
                    default='coco',
                    help='Dataset name',
                    choices=['textcap', 'coco', "sbucaptions", "combined", "yfcc7m"])
parser.add_argument(
    '--image_encoder_name',
    default='resnet50',
    choices=['resnet18', 'resnet50', "efficientnet"],
    help=
    'image model architecture: resnet18, resnet50 or efficientnet (default: resnet50)'
)
parser.add_argument(
    '--text_encoder_name',
    default='distilbert-base-uncased',
    choices=['distilbert-base-uncased'],
    help=
    'text model architecture: distilbert-base-uncased (default: distilbert-base-uncased)'
)
parser.add_argument('-save_model_dir',
                    default='./models',
                    help='Path where models')
parser.add_argument('--num_epochs',
                    default=2,
                    type=int,
                    help='Number of epochs for training')
parser.add_argument('--image_size', default=224, type=int, help='Image size')
parser.add_argument('-b',
                    '--batch_size',
                    default=256,
                    type=int,
                    help='Batch size')
parser.add_argument('-lr', '--learning_rate', default=1e-4, type=float)
parser.add_argument('-wd', '--weight_decay', default=0.0, type=float)
parser.add_argument('--fp16_precision',
                    action='store_true',
                    help='Whether to use 16-bit precision for GPU training')
parser.add_argument('--imagenet_eval',
                    action='store_true',
                    help='Whether to evaluate on imagenet validation dataset. Required huggingface imagenet-1k dataset.')
parser.add_argument('--imagenet_eval_steps',
                    default=1000,
                    type=int,
                    help='Evaluate on imagenet every N steps')
parser.add_argument('--log_every_n_steps',
                    default=50,
                    type=int,
                    help='Log every n steps')
parser.add_argument('--ckpt_path',
                    default=None,
                    type=str,
                    help='Specify path to clip_model.pth to resume training')


def main():
    args = parser.parse_args()
    train_clip(args)


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F

from simple_clip.utils import get_feature_size


def clip_loss(logits):
    targets = torch.arange(logits.size(0)).to(logits.device)
    loss_images = F.cross_entropy(logits, targets)
    loss_texts = F.cross_entropy(logits.t(), targets)

    return loss_images, loss_texts


class CLIP(torch.nn.Module):

    def __init__(self,
                 image_encoder,
                 text_encoder,
                 mlp_out_dim=768,
                 mlp_in_dim=False):
        super(CLIP, self).__init__()

        if not mlp_in_dim:
            mlp_in_dim = get_feature_size(image_encoder)

        self.image_encoder = image_encoder
        self.text_encoder = text_encoder

        self.projection_head = torch.nn.Sequential(
            torch.nn.Linear(mlp_in_dim, mlp_in_dim, bias=False),
            torch.nn.ReLU(),
            torch.nn.Linear(mlp_in_dim, mlp_out_dim, bias=False))

    def forward(self, image, input_ids, attention_mask):
        image_features = self.extract_image_features(image)
        text_features = self.extract_text_features(input_ids, attention_mask)
        # logits per image
        logits = image_features @ text_features.t()
        return logits

    def extract_image_features(self, images):
        image_features = self.image_encoder(images)
        return self.projection_head(image_features)

    def extract_text_features(self, input_ids, attention_mask):
        return self.text_encoder(input_ids, attention_mask)



import torch
from transformers import DistilBertTokenizer
import datasets
import webdataset as wds

from simple_clip.encoders import resnet18, resnet50, efficientnet_v2_s, TextEncoder
from simple_clip.custom_datasets.clip_datasets import COCODataset, SBUDataset, CombinedDataset


def get_dataset(dataset_name,
                dataset_path,
                transforms,
                split="train",
                shuffle_captions=True):
    tokenizer = DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
    if dataset_name == "coco":
        # TODO: combine textcap and coco
        data = datasets.load_dataset("MMInstruction/M3IT", dataset_name)[split]
        return COCODataset(data,
                           tokenizer,
                           transforms=transforms,
                           image_key="image_base64_str",
                           text_key="outputs",
                           shuffle_captions=shuffle_captions)
    elif dataset_name == "textcap":
        data = datasets.load_dataset("MMInstruction/M3IT", dataset_name)[split]
        return COCODataset(data,
                           tokenizer,
                           transforms=transforms,
                           image_key="image_base64_str",
                           text_key="outputs",
                           shuffle_captions=shuffle_captions)
    elif dataset_name == "sbucaptions":
        data = datasets.load_from_disk(f"{dataset_path}/sbu_captions_images")["train"]
        return SBUDataset(data,
                          tokenizer,
                          transforms=transforms,
                          image_key="image",
                          text_key="caption")
    elif dataset_name == "combined":
        data_coco = datasets.load_from_disk(f"{dataset_path}/coco")["train"]
        data_textcap = datasets.load_from_disk(f"{dataset_path}/textcap")["train"]
        data_sbu = datasets.load_from_disk(f"{dataset_path}/sbu_captions_images")
        data_sbu = data_sbu.map(lambda example: {"caption": [example["caption"]]})
        data = datasets.concatenate_datasets([data_coco, data_textcap, data_sbu])
        return CombinedDataset(data,
                               tokenizer,
                               transforms=transforms)
    elif dataset_name == "yfcc7m":
        def prepare_data(x):
            caption = x["txt"]
            encoded_caption =  tokenizer(caption,
                                padding="max_length",
                                truncation=True,
                                max_length=100)
            image = transforms(x["jpg"])
            
            instance = {
                key: torch.tensor(value)
                for key, value in encoded_caption.items()
            }
            instance["image"] = image
            
            return instance
        
        dataset = wds.WebDataset([f"{dataset_path}/yfcc7m/{i:05d}.tar" for i in range(1538)], shardshuffle=True, cache_dir=f"{dataset_path}/yfcc7m_training_cache")
        dataset = dataset.shuffle(1000, initial=100).decode("pil").map(prepare_data)
        return dataset
    raise Exception(f"Invalid dataset name {dataset_name} - options are [coco, sbucaptions, combined, yfcc7m]")


def get_image_encoder(model_name):
    if model_name == "resnet18":
        return resnet18()
    elif model_name == "resnet50":
        return resnet50()
    elif model_name == "efficientnet":
        return efficientnet_v2_s()
    raise Exception(
        "Invalid model name - options are [resnet18, resnet50, efficientnet]")


def get_text_encoder(model_name="distilbert-base-uncased"):
    return TextEncoder(model_name)


def accuracy(output, target, topk=(1, )):
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


@torch.inference_mode
def get_feature_size(encoder):
    """Get the feature size from the encoder using a dummy input."""
    encoder.eval()
    dummy_input = torch.randn(1, 3, 32, 32)
    output = encoder(dummy_input)
    return output.shape[1]



import datasets

import torch
from torch.utils.data import DataLoader, Dataset
from transformers import DistilBertTokenizer

from tqdm.auto import tqdm
from sklearn.metrics import accuracy_score


class ImageNetValidation:

    def __init__(self, transform):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        data = datasets.load_dataset("imagenet-1k")
        dataset = ImageNetDataset(data["validation"], transform)
        self.dataloader = DataLoader(dataset,
                                batch_size=256,
                                num_workers=4)
        
        labels = data["validation"].features["label"].int2str(list(range(1000)))
        self.label_queries = [f"a photo of a {l}" for l in labels]

        tokenizer = DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
        self.encoded_texts = tokenizer(self.label_queries, padding=True, truncation=True, max_length=100)

    @torch.inference_mode
    def evaluate(self, model):
        model.eval()
        with torch.no_grad():
            image_features, labels = self._get_image_embs_labels(model)
            text_features = self._get_text_embs(model)

            preds = image_features @ text_features.t()

            labels = labels.cpu().detach().tolist()
            
            preds = preds.argmax(dim=-1).cpu().detach().tolist()

            acc = accuracy_score(labels, preds)
            print("Accuracy ImagNet Val: ", acc)

    def _get_image_embs_labels(self, model):
        embs, labels = [], []
        for images, targets in tqdm(self.dataloader):
            with torch.no_grad():
                images = images.to(self.device)
                out = model.extract_image_features(images)
                features = out.cpu().detach().tolist()
                embs.extend(features)
                labels.extend(targets.cpu().detach().tolist())
        return torch.tensor(embs).to(self.device), torch.tensor(labels).to(self.device)
    
    def _get_text_embs(self, model):
        input_ids = torch.tensor(self.encoded_texts["input_ids"]).to(self.device)
        attention_mask = torch.tensor(self.encoded_texts["attention_mask"]).to(self.device)
        return model.extract_text_features(input_ids, attention_mask)
        

class ImageNetDataset(Dataset):

    def __init__(self,
                 data,
                 transforms):

        self.data = data
        self.transforms = transforms

    def __getitem__(self, idx):
        image = self.data[idx]["image"]
        label = self.data[idx]["label"]
        image = self.transforms(image)
        return image, torch.tensor(label)

    def __len__(self):
        return len(self.data)



import torch.nn as nn
from torchvision import models
from torchvision.models import ResNet50_Weights, ResNet18_Weights, EfficientNet_V2_S_Weights


class Squeeze(nn.Module):

    def forward(self, x):
        return x.squeeze(-1).squeeze(-1)


def _prep_encoder(model):
    modules = list(model.children())[:-1]
    modules.append(nn.AdaptiveAvgPool2d(1))
    modules.append(Squeeze())

    return nn.Sequential(*modules)


def resnet18():
    resnet = models.resnet18(weights=ResNet18_Weights.IMAGENET1K_V1)
    return _prep_encoder(resnet)


def resnet50():
    resnet = models.resnet50(weights=ResNet50_Weights.IMAGENET1K_V2)
    return _prep_encoder(resnet)


def efficientnet_v2_s():
    model = models.efficientnet_v2_s(
        weights=EfficientNet_V2_S_Weights.IMAGENET1K_V1)
    return _prep_encoder(model)



from setuptools import setup, find_packages
import os


def read_long_description():
    root = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(root, "README.md")
    with open(path, "r", encoding="utf-8") as f:
        text = f.read()
    return text


setup(
    name = 'simple-clip',
    packages = find_packages(exclude=['notebooks']),
    version = '0.1.0',
    license='MIT',
    description = 'A minimal, but effective implementation of CLIP (Contrastive Language-Image Pretraining) in PyTorch',
    author = 'Filip Basara',
    author_email = 'basarafilip@gmail.com',
    url = 'https://github.com/filipbasara0/simple-clip',
    long_description=read_long_description(),
    long_description_content_type = 'text/markdown',
    keywords = [
        'machine learning',
        'pytorch',
        'self-supervised learning',
        'representation learning',
        'contrastive learning'
    ],
    install_requires=[
        'torch>=2.1',
        'torchvision>=0.16',
        'transformers>=4.35',
        'datasets>=2.15',
        'tqdm>=4.66',
        'torchinfo>=1.8.0',
        'webdataset>=0.2.77',
        'scikit-learn>=1.3.2'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
    ],
    scripts=['run_training.py', 'train.py'],
    entry_points={
        "console_scripts": [
            "train_clip = run_training:main"
        ],
    },
)